## @file CurveADT.py
#  @author Henry Madej
#  @brief Provides the Abstract Data Type (ADT) for representing Curves
#  @date 07/01/2018

import numpy as np
import re
from SeqADT import *

def __get_points_from_sequence__(sequence, position, number_of_points):
  values = []
  for i in range(number_of_points):
    values.append(sequence.get(position+i))
  return values

## @brief An abstract data type that represents a Curve
class CurveT():
  __PATTERN__ = re.compile('(\d+.\d*|\d*.\d+|\d+),\s(\d+.\d*|\d*.\d+|\d+)')

  ## @brief CurveT constructor
  #  @details Initializes a CurveT object from a set of points provided in file,
  #   filename
  def __init__(self, filename):
    self.x_sequence = SeqT()
    self.y_sequence = SeqT()
    self.__read_file__(filename)


  def __read_file__(self, filename):
    file = open(filename, 'r')
    index = 0
    for line in file:
      match = self.__PATTERN__.match(line)
      if match:
        self.x_sequence.add(index, float(match.group(1)))
        self.y_sequence.add(index, float(match.group(2)))
        index += 1
    file.close()


  ## @brief linVal linear interpolation of the curve to approximate output y
  #         given some input x
  #  @param x the input value of x to the line to predict the output y
  #  @return y the output of the line at input x
  def linVal(self, x):
    if self.x_sequence.size() < 2:
        raise ValueError("Not enough points for linear interpolation")
    position_in_sequence = self.x_sequence.indexInSeq(x)
    x_ = __get_points_from_sequence__(self.x_sequence, position_in_sequence, 2)
    y_ = __get_points_from_sequence__(self.y_sequence, position_in_sequence, 2)
    denominator = x_[1] - x_[0]
    y = ((y_[1] - y_[0]) / denominator) * (x - x_[0]) + y_[0]
    return y

  ## @brief quadVal quadratic interpolation of the curve to approximate output
  #         y given some input x
  #  @param x the input value of x to the quadratic to predict output y
  #  @return y the output of the quadratic at input x
  def quadVal(self, x):
    if self.x_sequence.size() < 3:
      raise ValueError("Not enough points for quadratic interpolation")
    position_in_sequence = self.x_sequence.indexInSeq(x)
    xs = __get_points_from_sequence__(self.x_sequence, position_in_sequence, 3)
    ys = __get_points_from_sequence__(self.y_sequence, position_in_sequence, 3)
    b =  ((ys[2] - ys[0]) / (xs[2] - xs[0])) * (x - xs[1])
    a = ((ys[2] - 2 * ys[1] + ys[0]) / (2 * (xs[2] - xs[1]) ** 2)) * ((x - xs[1]) ** 2)
    return a + b + ys[1]


  ## @brief Regression of n degree polynomial to approximate output y given some
  #         input x
  #  @param n the highest degree of the polynomial to be fitted
  #  @param x the input value of x to the polynomial to predict output y
  #  @return y the output of the polynomial at input x
  def npolyVal(self, n, x):
    xs = np.array(__get_points_from_sequence__(self.x_sequence, 0, self.x_sequence.size()))
    ys = np.array(__get_points_from_sequence__(self.y_sequence, 0, self.y_sequence.size()))
    estimated_polynomial_function = np.poly1d(np.polyfit(xs, ys, n))
    return estimated_polynomial_function(x)
